//  $Id$
//      
//      I/O library functions for interfacing with SOLO
//      
//      relative path: ./lib/boot/m-pc
//  
//      Shag/OS: an Object-Oriented MicroKernel Operating System
//       Libraries
//        (boot)
//         (m-pc)
//  
//  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
//  
//  This file is a part of the Shag/OS operating system distribution.
//  
//  Shag/OS is free software for non-profit uses; it may be freely distributed
//  for non-profit purposes under the terms of the Shag/OS License, version 1.
//  
//  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
//  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
//  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
//  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
//  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
//  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
//  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
//  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
//  
//  You should have received a copy of the Shag/OS License along with this
//  program, in a file called 'COPYING'.  If this license was not received,
//  please contact the author and describe how this software was obtained.
//  
//  The author can be reached for questions, comments, and suggestions at:
//  
//      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
//      email:   shaggy@csh.rit.edu
//  
//      For bug reports, please read the file 'BUGS' first, then include the 
//      product name and version, and the word 'bug' in the subject.
//  
//  Complete, unaltered, original distributions of this and related code
//  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
//  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy

#pragma Make Includes_Banned += -nostdinc

#define USE_STDLIB
#include <sys/types.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <boot/fileinfo.h>
#include <boot/cpuinfo.h>
#include <boot/bootparam.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/fd.h>
#include <linux/hdreg.h>


// Imports...
typedef BootParam::index_t index_t;
typedef BootParam::fd_t fd_t;


BootParam *bootparam = 0;


void boot_reboot(bool cold)
{
	printf("EmSolo: reboot %s\n", cold ? "cold" : "warm");
	exit(1);
}


void boot_exit(int32 code)
{
	printf("EmSolo: exit 0x%x (%d)\n", code, code);
	exit(code);
}


void boot_panic(uint8 type, uint32 num, uint32 code, TssBlock *tss)
{
	printf("EmSolo: panic ");
	switch(type) {
	case PANIC_TYPE_IRQ:	printf("irq"); break;
	case PANIC_TYPE_EXC:	printf("exception"); break;
	default: 		printf("[%02x]", type);
	} 
	printf(" %d code=%#x tss=%p\n", num, code, tss); 
	exit(1);
}


void boot_callbios(uint8 vec, RegBlock * /*reg*/)
{
	printf("EmSolo: callbios (unimplemented)\n");
}


const MapFile* boot_readdir(index_t index)
{
	printf("EmSolo: readdir (unimplemented)\n");
	return 0;
}


void boot_seek(fd_t fd, uint32 ofs)
{
	if(lseek(fd, ofs, SEEK_SET) < 0)
		perror("lseek(,,SEEK_SET)");
}


uint32 boot_tell(fd_t fd)
{
uint32 ofs;
	if((ofs = lseek(fd, 0, SEEK_CUR)) == (uint32)-1) 
		perror("lseek(,,SEEK_CUR)");
	return ofs;
}


void boot_fstat(fd_t fd, FileInfo *info)
{
struct stat stbuf;
	fstat(fd, &stbuf);
	info->rdev = stbuf.st_rdev;
	info->atime = (int64)stbuf.st_atime*1000000000;
	info->ctime = (int64)stbuf.st_ctime*1000000000;
	info->mtime = (int64)stbuf.st_mtime*1000000000;
	info->size = stbuf.st_size;
	info->blksize = stbuf.st_blksize;
	info->blocks = stbuf.st_blocks;

	info->cylinders = 0;
	info->heads = 0;
	info->sectors = 0;

	switch(MAJOR(stbuf.st_rdev)) {
	case FLOPPY_MAJOR:
		struct floppy_struct fdgeo;
		if(!ioctl(fd, FDGETPRM, &fdgeo)) {
			info->cylinders = fdgeo.track;
			info->heads = fdgeo.head;
			info->sectors = fdgeo.sect;
		}
		break;
	case HD_MAJOR: case XT_DISK_MAJOR: case SCSI_DISK_MAJOR:
		struct hd_geometry hdgeo;
		if(!ioctl(fd, HDIO_GETGEO, &hdgeo)) { 
			info->cylinders = hdgeo.cylinders;
			info->heads = hdgeo.heads;
			info->sectors = hdgeo.sectors;
		}
		break;
	}
}


int32 boot_read(fd_t fd, void *adr, uint32 len)
{
	return read(fd, adr, len);
}


int32 boot_write(fd_t fd, const void *adr, uint32 len)
{
	return write(fd, adr, len);
}


fd_t boot_open(const char *name)
{
int fd;
	if((fd = open(name, O_RDONLY)) < 0)
		return (fd_t)0;
	else 
		return (fd_t)fd;
}


void boot_close(fd_t fd)
{
	close(fd);
}



extern "C" bool _bootparam_init()
{
static bool initialized = false;
static bool initializing = false;
static BootParam bootparambuf;
static char bootbuf[8192];
static char msgbuf[1024];
static struct CpuInfo cpuinfo;
	if(initialized)
		return true;
	if(initializing)
		return false;
	initializing = true;

	bootparam = &bootparambuf;

	bootparam->bootver_major = 0;
	bootparam->bootver_minor = 95;
	bootparam->bootname = "EmSOLO";
	bootparam->buffer = (uint32)bootbuf;
	bootparam->bufsize = sizeof(bootbuf);
	bootparam->bootdev = 0;		// invalid
	bootparam->entry = 0;		// unknown
	bootparam->lomem = (uint32)malloc(1*_M_);
	bootparam->himem = bootparam->lomem + 1*_M_;

	// bootparam->text, data, spare, bss
	// base_addr, base_size, base_lomem, base_himem
	// ext_addr, ext_size
	bootparam->msgbuf = (uint32)msgbuf;
	bootparam->msgsize = sizeof(msgbuf);
	bootparam->msgptr = 0;
	bootparam->cpuinfo = &cpuinfo;

	bootparam->console_fd = STDIN_FILENO;
	bootparam->self_fd = 0;		// invalid

	bootparam->exit = exit;
	bootparam->reboot = boot_reboot;
	bootparam->panic = boot_panic;
	bootparam->callbios = boot_callbios;
	bootparam->readdir = boot_readdir;
	bootparam->read = boot_read;
	bootparam->write = boot_write;
	bootparam->open = boot_open;
	bootparam->close = boot_close;
	bootparam->seek = boot_seek;
	bootparam->tell = boot_tell;
	bootparam->fstat = boot_fstat;

	initialized = true;
	initializing = false;
	return true;
}


bool _bootparam_initialized = _bootparam_init();


