/*  $Id$
 *      
 *      I/O library functions for interfacing with SOLO
 *      
 *      relative path: ./lib/boot/m-pc
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        (boot)
 *         (m-pc)
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution.
 *  
 *  Shag/OS is free software for non-profit uses; it may be freely distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'COPYING'.  If this license was not received,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For bug reports, please read the file 'BUGS' first, then include the 
 *      product name and version, and the word 'bug' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#include <types.h>
#include <posix/stdarg.h>

extern_C_begin

typedef int solo_fd_t;

struct FileInfo;


void solo_exit(int val);
char* solo_gets(char *s, size_t len);
int solo_puts(const char *s);
int solo_printf(const char *fmt, ...); 
int solo_vprintf(const char *fmt, va_list args);
solo_fd_t solo_open(const char *path, int flags);
int solo_close(solo_fd_t);
int solo_read(solo_fd_t, void *buf, size_t size);
int solo_write(solo_fd_t, const void *buf, size_t size);
off_t solo_lseek(solo_fd_t, off_t ofs, int whence);
void solo_fstat(solo_fd_t, struct FileInfo *info);
void* solo_malloc(size_t size);
void* solo_calloc(size_t num, size_t size);
void solo_free(void *ptr);
void* solo_realloc(void *ptr, size_t size);


#define SOLO_SEEK_SET	0
#define SOLO_SEEK_CUR	1
#define SOLO_SEEK_END	2

#define SOLO_O_RDONLY	0
#define SOLO_O_WRONLY	1
#define SOLO_O_RDWR	2

extern solo_fd_t stdin_fileno;
extern solo_fd_t stdout_fileno;
extern solo_fd_t stderr_fileno;


extern_C_end


#ifndef SOLO_NOSTDIO

typedef solo_fd_t fd_t;

/* #define'ing these names would be a bad idea, since it 
   would globally change them.  C++ programs would get
   slaughtered if they overloaded any of these names

   The following way works, but causes conflicts still
   if files such as 'stdio.h' or 'stdlib.h' are included...

void (* const exit)(int) = solo_exit;
char* (* const gets)(char *, size_t) = solo_gets;
int (* const puts)(const char *) = solo_puts;
int (* const printf)(const char *, ...) = solo_printf;
fd_t (* const open)(const char *, int) = solo_open;
int (* const close)(fd_t) = solo_close;
int (* const read)(fd_t, void *, size_t) = solo_read;
int (* const write)(fd_t, const void *, size_t) = solo_write;
off_t (* const lseek)(fd_t, off_t, int) = solo_lseek;

*/


static void exit(int code);
static char* gets(char *, size_t);
static int puts(const char *);
static int vprintf(const char *, va_list);
static int printf(const char *, ...);
static fd_t open(const char *, int);
static int close(fd_t);
static int read(fd_t, void *, size_t);
static int write(fd_t, const void *, size_t);
static off_t lseek(fd_t, off_t, int);


inline void exit(int code) { solo_exit(code); while(1); }
inline char* gets(char *s, size_t len) { return solo_gets(s, len); }
inline int puts(const char *s) { return solo_puts(s); }
inline int vprintf(const char *fmt, va_list args)
	{ return solo_vprintf(fmt, args); }

#ifdef _cplusplus
inline
#endif
int printf(const char *fmt, ...) { va_list args; int len;
	va_start(args, fmt);
	len = solo_vprintf(fmt,args);
	va_end(args);
	return len; }
inline fd_t open(const char *name, int mode)
	{ return solo_open(name, mode); }
inline int close(fd_t fd) { return solo_close(fd); }
inline int read(fd_t fd, void *p, size_t len)
	{ return solo_read(fd, p, len); }
inline int write(fd_t fd, const void *p, size_t len)
	{ return solo_write(fd, p, len); }
inline off_t lseek(fd_t fd, off_t ofs, int whence)
	{ return solo_lseek(fd, ofs, whence); }

#undef	SEEK_SET
#define SEEK_SET	SOLO_SEEK_SET
#undef	SEEK_CUR
#define SEEK_CUR	SOLO_SEEK_CUR
#undef	SEEK_END
#define	SEEK_END	SOLO_SEEK_END

#undef	O_RDONLY
#define	O_RDONLY	SOLO_O_RDONLY
#undef	O_WRONLY
#define O_WRONLY	SOLO_O_WRONLY
#undef	O_RDWR
#define	O_RDWR		SOLO_O_RDWR

#undef	STDIN_FILENO
#define	STDIN_FILENO	stdin_fileno
#undef	STDOUT_FILENO
#define STDOUT_FILENO	stdout_fileno
#undef	STDERR_FILENO
#define	STDERR_FILENO	stderr_fileno

#endif


#ifndef SOLO_NOSTDMM

static void* calloc(size_t, size_t);
static void* malloc(size_t);
static void free(void *);
static void* realloc(void *, size_t);


inline void* calloc(size_t num, size_t size)
	{ return solo_calloc(num, size); }
inline void* malloc(size_t size)
	{ return solo_malloc(size); }
inline void free(void *ptr)
	{ solo_free(ptr); }
inline void* realloc(void *ptr, size_t size)
	{ return solo_realloc(ptr, size); }

#endif

