//  $Id$
//      
//      SOLO memory management routines
//      
//      relative path: ./lib/boot/m-pc
//  
//      Shag/OS: an Object-Oriented MicroKernel Operating System
//       Libraries
//        (boot)
//         (m-pc)
//  
//  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
//  
//  This file is a part of the Shag/OS operating system distribution.
//  
//  Shag/OS is free software for non-profit uses; it may be freely distributed
//  for non-profit purposes under the terms of the Shag/OS License, version 1.
//  
//  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
//  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
//  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
//  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
//  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
//  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
//  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
//  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
//  
//  You should have received a copy of the Shag/OS License along with this
//  program, in a file called 'COPYING'.  If this license was not received,
//  please contact the author and describe how this software was obtained.
//  
//  The author can be reached for questions, comments, and suggestions at:
//  
//      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
//      email:   shaggy@csh.rit.edu
//  
//      For bug reports, please read the file 'BUGS' first, then include the 
//      product name and version, and the word 'bug' in the subject.
//  
//  Complete, unaltered, original distributions of this and related code
//  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
//  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy

//  $Log$


#include <defines>
#include <boot/bootparam.h>
#include <boot/fileinfo.h>
#include <posix/string.h>
#include <posix/stdarg.h>
#include <posix/stdio.h>
#include <macro.h>


#define SOLO_NOSTDMM
#include "libsolo.h"

extern BootParam *bootparam;


// Note:
//	these memory allocation routines probably don't need to
//	be very efficient, since they are only supposed to be used
//	by utilities that run once, then exit, so even if they waste
//	memory, it's only very temporary.
//	Any real program (such as an operating system kernel)
//	should have its own memory allocation routines.
//	These are just here for convenience...
//	so, once memory fragments with these routines, it might never rejoin...
//	Use sparingly... 

//	XXX - some consistency checks could be made on these memory
//	blocks to detect corruption--- such as checking that 
//	'next' pointers are always greater, and 'prev' pointers are
//	always less, as well as some checks on the sizes

//	XXX - all of this code depends upon the free list staying in
//	order, as well as the main 'freelist' structure being less
//	in address than any of the nodes allocated


struct MemBlock {
	vaddr_t	size;
	union {
		MemBlock *next;
		MemBlock *prev;
	} ptr;
};


static MemBlock freelist;


void* solo_calloc(size_t num, size_t size)
{
	void *ptr = solo_malloc(num*size);
	memset(ptr, 0, num*size);
	return ptr;
}


void* solo_malloc(size_t size)
{
	MemBlock *fp_prev = &freelist;
	MemBlock *fp;

	// assert(is_pow2(sizeof(MemBlock)));
	size = align(size, sizeof(MemBlock));

	while((fp = fp_prev->ptr.next) && size > fp->size)
		fp_prev = fp_prev->ptr.next;

	if(!fp)
		return 0;

	void *adr = (void*)((vaddr_t)fp + sizeof(MemBlock));
	size_t freesize = fp->size;

	if(fp->size - size > sizeof(MemBlock)) {
		MemBlock *newfp = (MemBlock*)((vaddr_t)adr + size);
		newfp->ptr.next = fp->ptr.next;
		newfp->size = freesize - (size+sizeof(MemBlock));
		fp_prev->ptr.next = newfp;
		fp->size = size;
	} else 
		fp_prev->ptr.next = fp->ptr.next;
	fp->ptr.prev = fp_prev;
	return adr;
}


void* solo_realloc(void *ptr, size_t size)
{
	MemBlock *fp = (MemBlock*)((vaddr_t)ptr - sizeof(MemBlock));
	if(fp->size >= size)
		return ptr;
	void *newp = solo_malloc(size);
	memcpy(newp, ptr, size);
	solo_free(ptr);
	return newp;
}


void solo_free(void *ptr)
{
	if(!ptr) 
		return;
	MemBlock *fp = (MemBlock*)((vaddr_t)ptr - sizeof(MemBlock));
	MemBlock *fp_prev = fp->ptr.prev;
	while(fp_prev->ptr.prev && fp_prev->ptr.prev < fp_prev)
		fp_prev = fp_prev->ptr.prev;
	while(fp_prev->ptr.next && fp_prev->ptr.next < fp)
		fp_prev = fp_prev->ptr.next;
	fp->ptr.next = fp_prev->ptr.next;
	fp_prev->ptr.next = fp;
	// XXX - should do block rejoining here, but we have
	// to be careful to fix any 'prev' pointers that may point
	// at the block header we're eliminating
}



extern "C" bool _bootparam_init();

extern "C" bool _solo_mm_init()
{
	static bool initialized = false;
	static bool initializing = false;

	if(initialized)
		return true;
	if(initializing)
		return false;
	initializing = true;

	_bootparam_init();

	freelist.ptr.next = (MemBlock*)bootparam->lomem;
	freelist.ptr.next->size =
		bootparam->himem-bootparam->lomem - sizeof(MemBlock);
	freelist.ptr.next->ptr.next = 0;
	
	initialized = true;
	initializing = false;
	return true;
}

// Global constructor:
bool _solo_mm_initialized = _solo_mm_init();

